﻿unit AnimationrigSpine;

interface

uses SysUtils,Classes,Contnrs,StrUtils,math,
     Animationrig,AnimationrigMisc,AnimationrigState,AnimationrigPhysicsSpring,
     uSpineCommon,uSpineAnimation,uSpineSkeleton,uSpineAttachment,uSpineSkin;

type
  TAnimrigStateSpineRenderer = class(TAnimrigStateRenderer)
    protected
      fBoneCacheList : TStringList;
      fstate : pspAnimationState;
      fskeleton : pspSkeleton;
      function ExtractWorldMatrix(bone:pspBone) : TPhysicsMatrix4;
    public
      const SaveVersion = 1;
      AntiScale : single = 1.0;

      constructor Create;
      destructor Destroy; override;
      procedure Clear; override;

      procedure Update(deltatime:double); override;

      procedure SetAnimation(trackindex,usetrack: integer; const basename: string; loop:boolean; addanim:boolean; trackmix:TAnimrigMisc.MixBlend; duration:single; seekmode:boolean); override;
      procedure SetTrackBlendAlpha(trackindex,usetrack: integer; alpha1, alpha2, alpha3, alpha4: single; exists:boolean); override;
      procedure SetTrackSeekTime(trackindex,usetrack:integer; time1, time2, time3, time4: single; exists:boolean); override;
      procedure SetMixDuration(duration:single); override;
      procedure SetSkin(const skinname:string); override;
      function  GetBoneMatrix(const bonename:string):TAnimrigMisc.Matrix4; override;

      property AnimationState : pspAnimationState read fstate write fstate;
      property Skeleton : pspSkeleton read fskeleton write fskeleton;

      procedure SaveToBinary(ms:TMemoryStream);

      class procedure SpineAnimationStateSaveToBinary(state:pspAnimationState; ms:TMemoryStream);
      class function  SpineAnimationStateLoadFormBinary(state:pspAnimationState; ms:TMemoryStream):boolean;
  end;

  TAnimrigStateSpinePhysicsRenderer = class(TAnimrigStateSpineRenderer)
    protected
    public
      procedure AddPhysicsTransform(const externalmatrix:TAnimrigMisc.Matrix4); override;
      procedure SetupPhysics(setting:TStringList); override;
      procedure ApplyPhysics; override;
      procedure AppeasePhysics; override;
      procedure AddPhysicsForce(const target:string; x,y,z : double); override;
  end;

implementation

{ TAnimrigStateRendererSpine }

procedure TAnimrigStateSpineRenderer.Clear;
begin
  inherited;

  fstate := nil;
  fskeleton := nil;
  fBoneCacheList.Clear;
end;

constructor TAnimrigStateSpineRenderer.Create; //(state: pspAnimationState; skeleton:pspSkeleton);
begin
  inherited ;

  fObjectType := anoSpine;
  fBoneCacheList := TStringList.Create;

  Clear;
end;

destructor TAnimrigStateSpineRenderer.Destroy;
begin
  fBoneCacheList.Free;

  inherited;
end;

function TAnimrigStateSpineRenderer.ExtractWorldMatrix(bone: pspBone): TPhysicsMatrix4;
begin
  result := TAnimrigMisc.IdentityMatrix;
  result._11 := bone.a;
  result._12 := bone.c;
  result._21 := bone.b;
  result._22 := bone.d;
  result._41 := bone.worldX * AntiScale;
  result._42 := bone.worldY * AntiScale;
end;

function TAnimrigStateSpineRenderer.GetBoneMatrix(const bonename: string): TAnimrigMisc.Matrix4;
var
  bone : pspBone;
  i : integer;
  cachehit : boolean;
begin
  result := TAnimrigMisc.IdentityMatrix;
  if fskeleton = nil then exit;
  if fstate = nil then exit;

  // cache
  bone := nil;
  cachehit := false;
  for i:=0 to fBoneCacheList.Count-1 do
    if bonename = fBoneCacheList.Strings[i] then
    begin
      bone := pspBone(fBoneCacheList.Objects[i]);
      cachehit := true;
      break;
    end;
  if bone = nil then bone := spSkeleton_findBone(fskeleton,bonename);

  if not(cachehit) then fBoneCacheList.AddObject(bonename,pointer(bone));

  if bone = nil then exit;
  result := ExtractWorldMatrix(bone);
end;

procedure TAnimrigStateSpineRenderer.SaveToBinary(ms: TMemoryStream);
begin
  SpineAnimationStateSaveToBinary(fstate,ms);
end;

procedure TAnimrigStateSpineRenderer.SetAnimation(trackindex,usetrack: integer; const basename: string; loop, addanim:boolean; trackmix:TAnimrigMisc.MixBlend; duration:single; seekmode:boolean);
  procedure inSetEmpty(trackindex:integer; dt:single);
  begin
    if dt < 0.0 then dt := fstate.data.defaultMix;
    if addanim then
      spAnimationState_addEmptyAnimation(fstate,trackindex , dt , 0.0)
    else
      spAnimationState_setEmptyAnimation(fstate,trackindex , dt);
  end;
  procedure inSetAnimation(trackindex:integer; const name1,name2:string; dt:single);
  var
    anim : pspAnimation;
    track : pspTrackEntry;
    l : integer;
  begin
    // search
    anim := spSkeletonData_findAnimation(fstate.data.skeletonData,name1);
    if anim = nil then anim := spSkeletonData_findAnimation(fstate.data.skeletonData,name2);

    if anim = nil then
    begin
      // set/add empty
      inSetEmpty(trackindex , dt);
    end
    else
    begin
      // set/add animation
      l := 0;
      if loop then l := 1;
      // set animation
      if addanim then
        track := spAnimationState_addAnimation(fstate,trackindex,anim , l,0.0)
      else
        track := spAnimationState_setAnimation(fstate,trackindex,anim , l);
      if track <> nil then
      begin
        // mixblend
        if track.mixBlend >= SP_MIX_BLEND_REPLACE then   // ignore SETUP FIRST
        begin
          case trackmix of
            anmReplace : track.mixBlend := SP_MIX_BLEND_REPLACE;
            anmAdd     : track.mixBlend := SP_MIX_BLEND_ADD;
          end;
        end;
        // duration
        if duration < 0.0 then
        begin
          // use spine default mix
        end
        else
        begin
          track.mixDuration := max(dt,0.0);
        end;
        // seekmode
        if seekmode then track.timeScale := 0.0
                    else track.timeScale := 1.0;
      end;
    end;
  end;
begin
  if fstate = nil then exit;

  // layertype = anlSpine1
  if usetrack = 1 then
  begin
    if basename = 'empty' then
    begin
      inSetEmpty(trackindex+0 , duration);
    end
    else
    begin
      inSetAnimation(trackindex+0 , basename,basename+'@up' , duration);
    end;
  end;

  // layertype = anlSpine4
  if usetrack = 4 then
  begin
    if basename = 'empty' then
    begin
      inSetEmpty(trackindex+0 , duration);
      inSetEmpty(trackindex+1 , duration);
      inSetEmpty(trackindex+2 , duration);
      inSetEmpty(trackindex+3 , duration);
    end
    else
    begin
      inSetAnimation(trackindex+0 , basename,basename+'@up' , duration);
      inSetAnimation(trackindex+1 , basename+'@right','' , duration);
      inSetAnimation(trackindex+2 , basename+'@down','' , duration);
      inSetAnimation(trackindex+3 , basename+'@left','' , duration);
    end;
  end;
end;

procedure TAnimrigStateSpineRenderer.Update(deltatime: double);
var
  fpumask : cardinal;
  changed : boolean;
begin
  if fstate = nil then exit;
  if fskeleton = nil then exit;

  changed := false;
  fpumask := Get8087CW;
  if (fpumask and $3f) <> $3f then
  begin
    math.SetExceptionMask([exInvalidOp,exDenormalized,exZeroDivide,exOverflow,exUnderflow,exPrecision]);   // = $3F
    changed := true;
  end;

  spAnimationState_update(fstate , deltatime);
  spAnimationState_apply(fstate , fskeleton);     // *zero div
  spSkeleton_updateWorldTransform(fskeleton);    //x,y transform after , next call spSkeleton_updateWorldTransform

  if changed then Set8087CW(fpumask);

  // physics
  AddPhysicsTransform(TAnimrigMisc.IdentityMatrix);
  UpdatePhysics(deltatime);
  ApplyPhysics;
end;

procedure TAnimrigStateSpineRenderer.SetTrackBlendAlpha(trackindex,usetrack: integer; alpha1, alpha2, alpha3, alpha4: single; exists:boolean);
  procedure inSet(trackindex:integer; alpha:single);
  var
    track : pspTrackEntry;
  begin
    track := spAnimationState_getCurrent(fstate,trackindex);
    if track = nil then exit;

    track.alpha := alpha;
  end;
begin
  if fstate = nil then exit;

  // exists enable layer flag
  // if pass track alpha , check flag and return function.
  //if not(exists) then exit;

  // 1 animation blend layer
  if usetrack = 1 then
  begin
    inSet(trackindex+0 , alpha1);  // up
  end;

  // 4 animations blend layer
  if usetrack = 4 then
  begin
    inSet(trackindex+0 , alpha1);  // up
    inSet(trackindex+1 , alpha2);  // right
    inSet(trackindex+2 , alpha3);  // down
    inSet(trackindex+3 , alpha4);  // left
  end;
end;

procedure TAnimrigStateSpineRenderer.SetTrackSeekTime(trackindex, usetrack: integer; time1, time2, time3, time4: single; exists: boolean);
  procedure inSet(trackindex:integer; time_:single);
  var
    track : pspTrackEntry;
  begin
    track := spAnimationState_getCurrent(fstate,trackindex);
    if track = nil then exit;

    track.trackTime := time_;
  end;
begin
  if fstate = nil then exit;

  // exists enable layer flag
  // if pass track alpha , check flag and return function.
  //if not(exists) then exit;

  // 1 animation seek time layer
  if usetrack = 1 then
  begin
    inSet(trackindex+0 , time1);  // up
  end;

  // 4 animations seek time layer
  if usetrack = 4 then
  begin
    inSet(trackindex+0 , time1);  // up
    inSet(trackindex+1 , time2);  // right
    inSet(trackindex+2 , time3);  // down
    inSet(trackindex+3 , time4);  // left
  end;
end;

class function TAnimrigStateSpineRenderer.SpineAnimationStateLoadFormBinary(state: pspAnimationState; ms: TMemoryStream): boolean;
var
  endpoint : int64;
  b : boolean;
begin
  result := false;

  endpoint := -1;
  try
    // head
    TAnimrigMisc.ReadHead(ms,'spinestate');
    // data size
    endpoint := TAnimrigMisc.ReadSkipBegin(ms);
    // state exists
    if state = nil then
    begin
      // skip
    end
    else
    begin
      // data exists
      b := TAnimrigMisc.ReadBool(ms);
      if b then
      begin
        // exists
      end
      else
      begin
        // null
      end;
    end;
    result := true;
  except
    //
  end;

  // end point
  TAnimrigMisc.JumpSkipEnd(ms,endpoint);
end;

class procedure TAnimrigStateSpineRenderer.SpineAnimationStateSaveToBinary(state: pspAnimationState; ms:TMemoryStream);
  procedure inTrack(track : pspTrackEntry);
  var
    endpoint : int64;
    s : string;
  begin
    // head
    TAnimrigMisc.WriteHead(ms , 'track');
    // data size
    endpoint := TAnimrigMisc.WriteSkipBegin(ms);

    // exists
    if track = nil then
    begin
      TAnimrigMisc.WriteBool(ms , false);
    end
    else
    begin
      TAnimrigMisc.WriteBool(ms , true);
      // animation (include <empty>)
      s := UTF8ToString(track.animation.name);
      TAnimrigMisc.WriteString(ms , s);
      // state
      TAnimrigMisc.WriteInteger(ms , track.loop);
      TAnimrigMisc.WriteFloat(ms , track.delay);
      TAnimrigMisc.WriteFloat(ms , track.trackTime);
      TAnimrigMisc.WriteFloat(ms , track.timeScale);
      TAnimrigMisc.WriteFloat(ms , track.alpha);
      TAnimrigMisc.WriteInteger(ms , integer(track.mixBlend));
      // next
      inTrack(pspTrackEntry(track.next));
    end;

    // write data size
    TAnimrigMisc.WriteSkipEnd(ms,endpoint);
  end;
var
  i : integer;
  endpoint : int64;
begin
  // head
  TAnimrigMisc.WriteHead(ms , 'spinestate');
  // data size
  endpoint := TAnimrigMisc.WriteSkipBegin(ms);

  // exists
  if state = nil then
  begin
    TAnimrigMisc.WriteBool(ms , false);
  end
  else
  begin
    TAnimrigMisc.WriteBool(ms , true);
    // version
    TAnimrigMisc.WriteInteger(ms , SaveVersion);
    // tracks
    TAnimrigMisc.WriteHead(ms , 'tracks');
    TAnimrigMisc.WriteInteger(ms , state.tracksCount);
    // track
    for i:=0 to state.tracksCount-1 do
    begin
      inTrack(state.tracks[i]);
    end;
  end;

  // write data size
  TAnimrigMisc.WriteSkipEnd(ms,endpoint);
end;

procedure TAnimrigStateSpineRenderer.SetMixDuration(duration: single);
begin
  if fstate = nil then exit;

  fstate.data.defaultMix := duration;
end;

procedure TAnimrigStateSpineRenderer.SetSkin(const skinname: string);
var
  skin : pspSkin;
begin
  if fskeleton = nil then exit;

  skin := spSkeletonData_findSkin(fskeleton.data,skinname);
  if skin <> nil then
  begin
    spSkeleton_setSkin(fskeleton,skin);
    spSkeleton_setToSetupPose(fskeleton);
  end;
end;

{ TAnimrigStateSpinePhysicsRenderer }

procedure TAnimrigStateSpinePhysicsRenderer.AddPhysicsForce(const target:string; x, y, z: double);
var
  p : pointer;
  g : TPhysicsGroup;
  v : TPhysicsVector3;
begin
  // safe check
  if fstate = nil then exit;
  if fskeleton = nil then exit;

  if target = 'none' then exit;

  v.x := x;
  v.y := y;
  v.z := z;
  for p in fPhysicsGroups do
  begin
    g := TPhysicsGroup(p);
    if (g.AliasName = target) or (target = 'all') then g.AddForce(v);
  end;
end;

procedure TAnimrigStateSpinePhysicsRenderer.AddPhysicsTransform(const externalmatrix:TAnimrigMisc.Matrix4);
var
  i : integer;
  g : TPhysicsGroup;
  m : TPhysicsMatrix4;
begin
  for i:=0 to fPhysicsGroups.Count-1 do
  begin
    g := TPhysicsGroup(fPhysicsGroups[i]);
    if g.GroupType = gtSpring then
    begin
      m := ExtractWorldMatrix( pspBone(TPhysicsGroupSpringBone(g).RootUserObject) );
      TAnimrigMisc.m4PushMatrix(m,TransMatrix);
      TPhysicsGroupSpringBone(g).RootMatrix := m;
      TPhysicsGroupSpringBone(g).ExternalMatrix := externalmatrix;
    end;
  end;
end;

procedure TAnimrigStateSpinePhysicsRenderer.AppeasePhysics;
var
  p : pointer;
  g : TPhysicsGroup;
begin
  // safe check
  if fstate = nil then exit;
  if fskeleton = nil then exit;

  for p in fPhysicsGroups do
  begin
    g := TPhysicsGroup(p);
    g.Appease;
  end;
end;

procedure TAnimrigStateSpinePhysicsRenderer.ApplyPhysics;
  procedure inUpdateWorldMatrix(bone:pspBone);
  var
    i : integer;
  begin
    if bone = nil then exit;
    spBone_updateWorldTransform(bone);

    for i:=0 to bone.childrenCount-1 do
    begin
      inUpdateWorldMatrix(pspBone(bone.children[i]));
    end;
  end;
  procedure inApplyBone(g:TPhysicsGroupSpringBone; skeleton:pspSkeleton);
  var
    p : pointer;
    b : TPhysicsNodeBone;
    sb : pspBone;
  begin
    for p in g.NodeList do
    begin
      b := TPhysicsNodeBone(p);
      if b.UserObject <> nil then
      begin
        sb := pspBone(b.UserObject);
        case g.Mode of
          gsmBone :
            begin
              sb.rotation := b.aLocalAngle * 360.0;
            end;
          gsmStrBone :
            begin
              sb.rotation := b.aLocalAngle * 360.0;
              sb.x := b.aLocalPosition.x / AntiScale;
              sb.y := b.aLocalPosition.y / AntiScale;
            end;
          gsmStrPoint,gsmPoint :
            begin
              sb.x := b.aLocalPosition.x / AntiScale;
              sb.y := b.aLocalPosition.y / AntiScale;
            end;
        end;
      end;
      // update world matrix , all children
      //inUpdateWorldMatrix(pspBone(g.RootBone.UserObject));
    end;
  end;
var
  p : Pointer;
  g : TPhysicsGroup;
begin
  // safe check
  if fstate = nil then exit;
  if fskeleton = nil then exit;

  for p in fPhysicsGroups do
  begin
    g := TPhysicsGroup(p);
    g.Apply;
    if g.GroupType = gtSpring then inApplyBone(TPhysicsGroupSpringBone(g),fskeleton);
  end;

  spSkeleton_updateWorldTransform(fskeleton);
end;

procedure TAnimrigStateSpinePhysicsRenderer.SetupPhysics(setting:TStringList);
  // get setting value
  function inGetSettingFloat(index:integer; const n:string; def_,min_,max_:double):double;
  var
    sl : TStringList;
    s : string;
  begin
    sl := TAnimrigMisc.DivideString(',',setting.Values[n]);
    s := '';
    if index < sl.Count then s := sl.Strings[index];
    sl.Free;

    result := StrToFloatDef(s,def_);
    result := min(max(result,min_),max_);
  end;
  function inGetSettingString(const n:string; const def_:string):string;
  begin
    result := setting.Values[n];
    if result = '' then result := def_;
  end;
  // bone
  procedure inAddBone(g:TPhysicsGroupSpringBone; bone:pspBone; parent:TPhysicsNodeBone);
  var
    b : TPhysicsNodeBone;
    s : string;
    i : integer;
  begin
    // terminate code
    if parent <> nil then
    begin
      s := UTF8ToString(bone.data.name);
      if pos('@end',s) > 0 then exit;    // terminate code
      if pos('@',s) > 0 then exit;       // when other code exists, safe exit.
    end;

    b := TPhysicsNodeBone.Create(parent);
    b.UserObject := bone;
    b.LocalAngle := bone.data.rotation / 360.0;
    b.LocalLength := bone.data.length * AntiScale;
    b.LocalPosition.x := bone.data.x * AntiScale;
    b.LocalPosition.y := bone.data.y * AntiScale;

    g.AddBone(b);

    for i:=0 to bone.childrenCount-1 do
      inAddBone(g,pspBone(bone.children[i]),b);

    // create terminate bone
    if bone.childrenCount = 0 then
    begin
      parent := b;
      b := TPhysicsNodeBone.Create(parent);
      b.UserObject := nil;
      b.LocalAngle := 0;
      b.LocalLength := 100;
      b.LocalPosition.x := parent.LocalLength;
      b.LocalPosition.y := 0;
      g.AddBone(b);
    end;
  end;
  // spring
  procedure inAddSpring(g:TPhysicsGroupSpringBone; bone:TPhysicsNodeBone);
  var
    p : pointer;
    s : TPhysicsSpringBone;
  begin
    for p in bone.Children do
    begin
      s := TPhysicsSpringBone.Create(g,bone,TPhysicsNodeBone(p));
      g.AddSpring(s);
      inAddSpring(g,TPhysicsNodeBone(p));
    end;
  end;
  // spring group
  procedure inAddSpringGroup(name,alias:string; bone : pspBone);
  var
    g : TPhysicsGroupSpringBone;
    root : pspbone;
    rootmatrix : TPhysicsMatrix4;
    s : string;
  begin
    // need has parent
    root := pspBone(bone.parent);
    if root = nil then exit;

    if alias = '' then alias := name;

    // enable / disable
    if inGetSettingString(alias+'.enable','true') = 'false' then exit;

    g := TPhysicsGroupSpringBone.Create;
    g.Name := name;
    g.AliasName := alias;
    g.RootUserObject := root;

    s := inGetSettingString(alias+'.mode','bone');
    if s = 'strpoint' then g.Mode := gsmStrPoint;
    if s = 'point' then g.Mode := gsmPoint;
    if s = 'strbone' then g.Mode := gsmStrBone;
    if s = 'bone' then g.Mode := gsmBone;

    g.K := inGetSettingFloat(0,alias+'.k',g.K , 0.1,100*100);
    g.Damping := inGetSettingFloat(0,alias+'.damping',g.Damping , 0.1,100*100);
    g.Mass := inGetSettingFloat(0,alias+'.mass',g.Mass , 0.001,1000);
    g.LimitVelocity := inGetSettingFloat(0,alias+'.limitvelocity',g.LimitVelocity, 1,$ffff);   // 100 = physics frame
    g.Gravity := inGetSettingFloat(0,alias+'.gravity',g.Gravity , 0,10000*100);
    g.GravityVector.x := inGetSettingFloat(0,alias+'.gravityvector',g.GravityVector.x , -1000,+1000);
    g.GravityVector.y := inGetSettingFloat(1,alias+'.gravityvector',g.GravityVector.y , -1000,+1000);
    g.GravityVector.z := inGetSettingFloat(2,alias+'.gravityvector',g.GravityVector.z , -1000,+1000);
    g.GravityVector.Normalize;
    g.LimitAngleMin := ingetSettingFloat(0,alias+'.limitangle',g.LimitAngleMin , -$ff,0);
    g.LimitAngleMax := ingetSettingFloat(1,alias+'.limitangle',g.LimitAngleMax , 0,+$ff);
    g.LimitMoveRatio := ingetSettingFloat(0,alias+'.limitmove',g.LimitMoveRatio , 0,+$ff);

    g.ForceRotateRatio := ingetSettingFloat(0,alias+'.forcerotate',g.ForceRotateRatio , 0.0,10.0);
    g.ForceRatio := ingetSettingFloat(0,alias+'.force',g.ForceRatio , 0.0,10.0);

    inAddBone(g,bone,nil);
    inAddSpring(g,g.RootBone);

    g.RootMatrix := ExtractWorldMatrix(root);
    g.SetupNatural;

    fPhysicsGroups.Add(g);
  end;
var
  i : integer;
  s,name,op,alias : string;
  idx : integer;
  sl : TStringList;
begin
  // do not forget to clear PhysicsGroups
  inherited ;

  if fstate = nil then exit;
  if fskeleton = nil then exit;

  spSkeleton_updateWorldTransform(fskeleton);

  for i:=0 to fskeleton.bonesCount-1 do
  begin
    s := UTF8ToString(fskeleton.bones[i].data.name);

    name := '';
    idx := pos('@',s);
    if idx = 0 then continue;

    // name @ spring,aliasname
    op := '';
    alias := '';
    name := copy(s,1,idx-1);
    sl := TAnimrigMisc.DivideString(',' , copy(s,idx+1,4096));
    if sl.Count > 0 then op := sl.Strings[0];
    if sl.Count > 1 then alias := sl.Strings[1];
    sl.Free;

    if op = 'spring' then inAddSpringGroup(name,alias,fskeleton.bones[i]);
  end;
end;

end.
